import json
import re
from typing import Optional, Union, Dict, Callable
import textwrap

from ...ai_api import AIAPI
from ..litellm_api import LiteLLMAIAPI
from .prompt_type import PromptType

from libbs.artifacts import Comment, Function, StackVariable
from jinja2 import Template, StrictUndefined

JSON_REGEX = re.compile(r"\{.*?}", flags=re.DOTALL)


class Prompt:
    DECOMP_REPLACEMENT_LABEL = "<DECOMPILATION>"
    SNIPPET_REPLACEMENT_LABEL = "<SNIPPET>"
    SNIPPET_TEXT = f"\n\"\"\"{SNIPPET_REPLACEMENT_LABEL}\"\"\""
    DECOMP_TEXT = f"\n\"\"\"{DECOMP_REPLACEMENT_LABEL}\"\"\""

    def __init__(
        self,
        name: str,
        template_name: str,
        desc: str = None,
        pretext_response: Optional[str] = None,
        posttext_response: Optional[str] = None,
        json_response: bool = True,
        response_key: str = None,
        ai_api=None,
        # callback(result, function, ai_api)
        gui_result_callback: Optional[Callable] = None
    ):
        self.name = name
        self.template_name = template_name
        self.last_rendered_template = None
        self._pretext_response = pretext_response
        self._posttext_response = posttext_response
        self._json_response = json_response
        self._response_key = response_key
        self._gui_result_callback = gui_result_callback
        self.desc = desc or name
        self.ai_api: LiteLLMAIAPI = ai_api

    def _load_template(self, prompt_style: PromptType) -> Template:
        from . import get_prompt_template
        template_text = get_prompt_template(self.template_name, prompt_style)
        return Template(textwrap.dedent(template_text), undefined=StrictUndefined)

    def query_model(self, *args, function=None, dec_text=None, use_dec=True, **kwargs):
        if self.ai_api is None:
            raise Exception("api must be set before querying!")

        @AIAPI.requires_function
        def _query_model(ai_api=self.ai_api, function=function, dec_text=dec_text, **_kwargs) -> Union[Dict, str]:
            if not ai_api:
                return {}

            ai_api.info(f"Querying {self.name} prompt with function {function}...")
            response = self._pretext_response if self._pretext_response and not self._json_response else ""
            template = self._load_template(self.ai_api.prompt_style)
            # grab decompilation and replace it in the prompt, make sure to fix the decompilation for token max
            query_text = template.render(
                decompilation=LiteLLMAIAPI.fit_decompilation_to_token_max(dec_text) if self.ai_api.fit_to_tokens else dec_text,
                few_shot=bool(self.ai_api.prompt_style == PromptType.FEW_SHOT),
            )
            self.last_rendered_template = query_text
            #ai_api.info(f"Prompting using model: {self.ai_api.model}...")
            #ai_api.info(f"Prompting with style: {self.ai_api.prompt_style}...")
            #ai_api.info(f"Prompting with: {query_text}")

            ai_api.on_query(self.name, self.ai_api.model, self.ai_api.prompt_style, function, dec_text)
            response += self.ai_api.query_model(query_text)
            #ai_api.info(f"Response received from AI: {response}")
            default_response = {} if self._json_response else ""
            if not response:
                ai_api.warning(f"Response received from AI was empty! AI failed to answer.")
                return default_response

            # changes response type to a dict
            if self._json_response:
                # if the response of OpenAI gets cut off, we have an incomplete JSON
                if "}" not in response:
                    response += "}"

                json_matches = JSON_REGEX.findall(response)
                if not json_matches:
                    return default_response

                json_data = json_matches[-1]
                try:
                    response = json.loads(json_data)
                except Exception:
                    response = {}

                if self._response_key is not None:
                    response = response.get(self._response_key, "")
            else:
                response += self._posttext_response if self._pretext_response else ""

            if isinstance(response, dict) or isinstance(response, str):
                resp_len = len(response)
                if resp_len:
                    ai_api.info(f"Response of len={resp_len} received from AI...")
                else:
                    ai_api.warning(f"Response recieved from AI, but it was empty! AI failed to answer.")
            else:
                ai_api.info("Reponse received from AI!")

            if ai_api.has_decompiler_gui and response:
                ai_api.info("Updating the decompiler with the AI response...")
                self._gui_result_callback(response, function, ai_api)

            return response
        return _query_model(ai_api=self.ai_api, function=function, dec_text=dec_text, use_dec=use_dec)

    @staticmethod
    def rename_function(result, function, ai_api: "AIAPI"):
        if function.name in result:
            new_name = result[function.name]
        else:
            new_name = list(result.values())[0]

        new_func = Function(name=new_name, addr=function.addr)
        ai_api._dec_interface.functions[function.addr] = new_func

    @staticmethod
    def rename_variables(result, function, ai_api: "AIAPI"):
        new_func: Function = function.copy()
        # clear out changes that are not for variables
        new_func.name = None
        new_func.type = None
        ai_api._dec_interface.rename_local_variables_by_names(function, result)

    @staticmethod
    def comment_function(result, function, ai_api: "AIAPI"):
        curr_cmt_obj = ai_api._dec_interface.comments.get(function.addr, None)
        curr_cmt = curr_cmt_obj.comment + "\n" if curr_cmt_obj is not None else ""

        ai_api._dec_interface.comments[function.addr] = Comment(
            addr=function.addr,
            comment=curr_cmt + result,
            func_addr=function.addr
        )



import logging
from typing import Dict

from binsync.data import Function, StackVariable, Comment, State, FunctionHeader

from dailalib.api import LiteLLMAIAPI
from dailalib.binsync_plugin.ai_bs_user import AIBSUser

_l = logging.getLogger(__name__)


class OpenAIBSUser(AIBSUser):
    DEFAULT_USERNAME = "chatgpt_user"

    def __init__(self, openai_api_key,  *args, **kwargs):
        super().__init__(openai_api_key, *args, **kwargs)
        self.ai_interface = LiteLLMAIAPI(openai_api_key=openai_api_key, decompiler_controller=self.controller, model=self._model)

    def run_all_ai_commands_for_dec(self, decompilation: str, func: Function, state: State):
        changes = 0
        artifact_edit_cmds = {
            self.ai_interface.RETYPE_VARS_CMD, self.ai_interface.RENAME_VARS_CMD, self.ai_interface.RENAME_FUNCS_CMD,
            self.ai_interface.ANSWER_QUESTION_CMD
        }
        cmt_prepends = {
            self.ai_interface.SUMMARIZE_CMD: "==== AI Summarization ====\n",
            self.ai_interface.ID_SOURCE_CMD: "==== AI Source Guess ====\n",
            self.ai_interface.FIND_VULN_CMD: "==== AI Vuln Guess ====\n",
        }

        func_cmt = ""
        new_func = Function(func.addr, func.size, header=FunctionHeader("", func.addr, args={}), stack_vars={})
        for cmd in self.ai_interface.AI_COMMANDS:
            # TODO: make this more explicit and change what is run
            if cmd not in {
                self.ai_interface.RENAME_FUNCS_CMD,
                self.ai_interface.ID_SOURCE_CMD,
                self.ai_interface.SUMMARIZE_CMD,
            }:
                continue

            try:
                resp = self.ai_interface.query_for_cmd(cmd, decompilation=decompilation)
            except Exception as e:
                _l.error(f"Failed to query for cmd {cmd} with error {e}")
                continue

            if not resp:
                continue

            if cmd not in artifact_edit_cmds:
                if cmd == self.ai_interface.ID_SOURCE_CMD:
                    if "http" not in resp:
                        continue

                func_cmt += cmt_prepends.get(cmd, "") + resp + "\n"
                # fake the comment actually being added to decomp
                decompilation = f"/* {Comment.linewrap_comment(resp)} */\n" + decompilation
                changes += 1

            elif cmd == self.ai_interface.RENAME_VARS_CMD:
                all_names = set(sv.name for _, sv in func.stack_vars.items())
                for off, sv in func.stack_vars.items():
                    old_name = sv.name
                    if old_name in resp:
                        proposed_name = resp[old_name]
                        if not proposed_name or proposed_name == old_name or proposed_name in all_names:
                            continue

                        if off not in new_func.stack_vars:
                            new_func.stack_vars[off] = StackVariable(sv.offset, "", None, func.stack_vars[off].size, func.addr)

                        new_func.stack_vars[off].name = proposed_name
                        decompilation = decompilation.replace(old_name, proposed_name)
                        changes += 1

            elif cmd == self.ai_interface.RETYPE_VARS_CMD:
                for off, sv in func.stack_vars.items():
                    old_name = sv.name
                    if old_name in resp:
                        proposed_type = resp[old_name]
                        if not proposed_type or proposed_type == sv.type:
                            continue

                        if off not in new_func.stack_vars:
                            new_func.stack_vars[off] = StackVariable(sv.offset, "", None, func.stack_vars[off].size, func.addr)

                        new_func.stack_vars[off].type = proposed_type
                        # we dont update decompilation here because it would be too weird
                        changes += 1

            elif cmd == self.ai_interface.RENAME_FUNCS_CMD:
                if func.name in resp:
                    proposed_name = resp[func.name]
                    if proposed_name in self.controller.functions() or not proposed_name or proposed_name == func.name:
                        continue

                    new_func.name = proposed_name
                    _l.info(f"Proposing new name for function {func.name} to {proposed_name}")
                    changes += 1

            elif cmd == self.ai_interface.ANSWER_QUESTION_CMD:
                answers: Dict[str, str] = resp
                current_cmts = state.get_func_comments(func.addr)
                for question, answer in answers.items():
                    for _, current_cmt in current_cmts.items():
                        if question in current_cmt.comment:
                            current_cmt.comment += f"\n{answer}"
                            state.set_comment(current_cmt)
                            changes += 1
                            break

        if changes:
            _l.info(f"Suggesting updates to {func} with diff: {new_func}")
            state.set_function(new_func)

        # send full function comment
        if func_cmt:
            state.set_comment(Comment(new_func.addr, func_cmt, func_addr=new_func.addr, decompiled=True), append=True)
            #self.controller.push_artifact(Comment(new_func.addr, func_cmt, func_addr=new_func.addr, decompiled=True), append=True)
            #self.controller.fill_comment(new_func.addr, user=self.username, artifact=Comment(new_func.addr, func_cmt, func_addr=new_func.addr, decompiled=True), append=True)
            #self.controller.schedule_job(
            #    self.controller.push_artifact,
            #    Comment(new_func.addr, func_cmt, func_addr=new_func.addr, decompiled=True),
            #    blocking=False,
            #    append=True
            #)

        return changes



RENAME_FUNCTION = """
# Task
You are a decompiled C expert that renames functions. When given a function, you rename it according to the
meaning of the function or its use. You specify which function you are renaming by its name.

You eventually respond with a valid json. As an example:
## Answer 
{
    "sub_404000": "fibonacci",
}

{% if (few_shot or cot) %}
# Example
Here is an example. Given the following code:
```
int sub_404000(int a0, char** a1)
{
    int is_even; // rax

    is_even = sub_404100(a0[1]) % 2 == 0
    return is_even;
}
```

You respond with:
## Answer
{
    "sub_404000": "is_even",
    "sub_404100": "get_value",
}
{% endif %}

# Example
Given the following code:
```
{{ decompilation }}
```

You respond with:
"""

RENAME_VARIABLES = """
# Task
You are decompiled C expert that renames variables in code. When given code, you rename variables according to the
meaning of the function or its use.

You eventually respond with a valid json. As an example:
## Answer 
{
  "v1": "i",
  "v2": "ptr"
}

{% if few_shot %}
# Example
Here is an example. Given the following code:
```
int sub_404000(int a0, char** a1)
{
    int v1; // rax

    v1 = sub_404100(a0[1]) % 2 == 0
    return v1;
}
```

You responded with:
## Answer
{
    "a0": "argc",
    "a1": "argv",
    "v1": "is_even"
}
{% endif %}

# Example
Given the following code:
```
{{ decompilation }}
```

You respond with:
"""

SUMMARIZE_FUNCTION = """
# Task
You are decompiled C expert that summarizes code. When given code, you summarize at a high level what the function does
and you identify if known algorithms are used in the function.

You eventually respond with a valid json. As an example:
## Answer 
{
    "summary": "This function computes the fibonacci sequence. It takes an integer as an argument and returns the fibonacci number at that index.",
    "algorithms": ["fibonacci"]
}

{% if few_shot %}
# Example
Here is an example. Given the following code:
```
int sub_404000(int a0, char** a1)
{
    int v1; // rax
    v1 = sub_404100(a0[1]) % 2 == 0
    return v1;
}
```

You responded with:
## Answer
{
    "summary": "This function takes two arguments and implements the is_even check on second argument",
    "algorithms": ["is_even"]
}
{% endif %}

# Example
Given the following code:
```
{{ decompilation }}
```

You respond with:
"""

IDENTIFY_SOURCE = """
# Task
You are a decompiled C expert that identifies the original source given decompilation. Upon discovering the source,
you give a link to the code.

You eventually respond with a valid json. As an example:
## Answer 
{
  "link": "https://github.com/torvalds/linux"
  "version": "5.10"
}

{% if few_shot %}
# Example
Here is an example. Given the following code:
```
void __fastcall __noreturn usage(int status)
{
    v2 = program_name;
    if ( status )
    {
        v3 = dcgettext(0LL, "Try '%s --help' for more information.\n", 5);
        _fprintf_chk(stderr, 1LL, v3, v2);
    }
    // ...
}
```

You would respond with:
## Answer
{
    "link": "https://www.gnu.org/software/coreutils/"
    "version": ""
}
{% endif %}

# Example
Given the following code:
```
{{ decompilation }}
```

You respond with:
"""


import logging
import os
import shutil
from pathlib import Path
import tempfile
from typing import Dict
import math
import threading

from binsync.api import load_decompiler_controller, BSController
from binsync.data.state import State
from binsync.data import (
    Function, Comment, StackVariable
)
from binsync.ui.qt_objects import (
    QDialog, QMessageBox
)

from dailalib.api import LiteLLMAIAPI
from tqdm import tqdm

_l = logging.getLogger(__name__)
_l.setLevel(logging.INFO)


class AIBSUser:
    MAX_FUNC_SIZE = 0xffff
    MIN_FUNC_SIZE = 0xff
    DEFAULT_USERNAME = "ai_user"

    def __init__(
        self,
        openai_api_key: str,
        binary_path: Path,
        bs_proj_path: Path = None,
        username: str = DEFAULT_USERNAME,
        copy_project=True,
        decompiler_backend=None,
        base_on=None,
        controller=None,
        model=None,
        progress_callback=None,
        range_str="",
        arch=None
    ):
        self._base_on = base_on
        self.username = username
        self._model = model
        self._progress_callback = progress_callback
        if bs_proj_path is not None:
            bs_proj_path = Path(bs_proj_path)

        # compute the range
        if range_str:
            range_strings = range_str.split("-")
            self.analysis_min = int(range_strings[0], 0)
            self.analysis_max = int(range_strings[1], 0)
        else:
            self.analysis_max = None
            self.analysis_min = None

        # copy or create the project path into the temp dir
        self.decompiler_backend = decompiler_backend
        self.project_path = bs_proj_path or Path(binary_path).with_name(f"{binary_path.with_suffix('').name}.bsproj")
        self._is_tmp = False

        self._on_main_thread = True if self.decompiler_backend is None else False
        if copy_project and self.project_path.exists():
            proj_dir = Path(tempfile.mkdtemp())
            shutil.copytree(self.project_path, proj_dir / self.project_path.name)
            self.project_path = proj_dir / self.project_path.name
            self._is_tmp = True

        create = False
        if not self.project_path.exists():
            create = True
            os.mkdir(self.project_path)

        # connect the controller to a GitClient
        _l.info(f"AI User working on copied project at: {self.project_path}")
        self.controller: BSController = load_decompiler_controller(
            force_decompiler=self.decompiler_backend, headless=True, binary_path=binary_path, callback_on_push=False,
            arch=arch
        )
        self.controller.connect(username, str(self.project_path), init_repo=create, single_thread=True)
        self.comments = {}

    def add_ai_user_to_project(self):
        # base all changes on another user's state
        if self._base_on:
            _l.info(f"Basing all AI changes on user {self._base_on}...")
            master_state = self.controller.get_state(user=self._base_on)
            master_state.user = self.username
        else:
            _l.info("Basing AI on current decompiler changes...")
            master_state = self.controller.get_state()

        # collect decompiled functions
        decompiled_functions = self._collect_decompiled_functions()
        #t = threading.Thread(
        #    target=self._query_and_commit_changes,
        #    args=(master_state, decompiled_functions,)
        #)
        #t.daemon = True
        #t.start()
        self._query_and_commit_changes(master_state, decompiled_functions)


    def _collect_decompiled_functions(self) -> Dict:
        valid_funcs = [
            addr
            for addr, func in self.controller.functions().items()
            if self._function_is_large_enough(func)
        ]

        if not valid_funcs:
            _l.info("No functions with valid size (small or big), to work on...")
            return {}

        # open a loading bar for progress updates
        #pbar = QProgressBarDialog(label_text=f"Decompiling {len(valid_funcs)} functions...")
        #pbar.show()
        #self._progress_callback = pbar.update_progress

        # decompile important functions first
        decompiled_functions = {}
        update_amt_per_func = math.ceil(100 / len(valid_funcs))
        callback_stub = self._progress_callback if self._progress_callback is not None else lambda x: x
        for func_addr in tqdm(valid_funcs, desc=f"Decompiling {len(valid_funcs)} functions for analysis..."):
            # bound check funcs size
            if ((self.analysis_max is not None and func_addr > self.analysis_max) or
                    (self.analysis_min is not None and func_addr < self.analysis_min)):
                callback_stub(update_amt_per_func)
                continue

            try:
                func = self.controller.function(func_addr)
            except Exception as e:
                _l.warning(f"Failed to get BS function at {hex(func_addr)}: {e}")
                func = None

            if func is None:
                callback_stub(update_amt_per_func)
                continue

            decompilation = self.controller.decompile(func_addr)
            if not decompilation:
                callback_stub(update_amt_per_func)
                continue

            decompiled_functions[func.addr] = (LiteLLMAIAPI.fit_decompilation_to_token_max(decompilation), func)
            callback_stub(update_amt_per_func)

        return decompiled_functions

    def _query_and_commit_changes(self, state, decompiled_functions):
        total_ai_changes = self.commit_ai_changes_to_state(state, decompiled_functions)
        if total_ai_changes:
            self.controller.client.commit_state(state, msg="AI initiated change to full state")
            self.controller.client.push()

        _l.info(f"Pushed {total_ai_changes} AI initiated changes to user {self.username}")

    def _function_is_large_enough(self, func: Function):
        return self.MIN_FUNC_SIZE <= func.size <= self.MAX_FUNC_SIZE

    def commit_ai_changes_to_state(self, state: State, decompiled_functions):
        ai_initiated_changes = 0
        update_cnt = 0
        round_updates = 0
        for func_addr, (decompilation, func) in tqdm(decompiled_functions.items(), desc=f"Querying AI for {len(decompiled_functions)} funcs..."):
            round_changes = self.run_all_ai_commands_for_dec(decompilation, func, state)
            ai_initiated_changes += round_changes
            if ai_initiated_changes:
                update_cnt += 1
                round_updates += round_changes

            if update_cnt >= 1:
                update_cnt = 0
                self.controller.client.commit_state(state, msg="AI Initiated change to functions")
                self.controller.client.push()
                _l.info(f"Pushed {round_updates} changes to user {self.username}...")
                round_updates = 0

        return ai_initiated_changes

    def run_all_ai_commands_for_dec(self, decompilation: str, func: Function, state: State):
        return 0



from .prompt_type import PromptType, DEFAULT_STYLE, ALL_STYLES
from .prompt import Prompt


class PromptNames:
    RENAME_FUNC = "RENAME_FUNCTION"
    RENAME_VARS = "RENAME_VARIABLES"
    SUMMARIZE_FUNC = "SUMMARIZE_FUNCTION"
    ID_SRC = "IDENTIFY_SOURCE"


def get_prompt_template(prompt_name, prompt_style):
    if prompt_style in [PromptType.FEW_SHOT, PromptType.ZERO_SHOT]:
        from .few_shot_prompts import RENAME_FUNCTION, RENAME_VARIABLES, SUMMARIZE_FUNCTION, IDENTIFY_SOURCE
        d = {
            PromptNames.RENAME_FUNC: RENAME_FUNCTION,
            PromptNames.RENAME_VARS: RENAME_VARIABLES,
            PromptNames.SUMMARIZE_FUNC: SUMMARIZE_FUNCTION,
            PromptNames.ID_SRC: IDENTIFY_SOURCE
        }
    elif prompt_style == PromptType.COT:
        from .cot_prompts import RENAME_FUNCTION, RENAME_VARIABLES, SUMMARIZE_FUNCTION, IDENTIFY_SOURCE
        d = {
            PromptNames.RENAME_FUNC: RENAME_FUNCTION,
            PromptNames.RENAME_VARS: RENAME_VARIABLES,
            PromptNames.SUMMARIZE_FUNC: SUMMARIZE_FUNCTION,
            PromptNames.ID_SRC: IDENTIFY_SOURCE
        }
    else:
        raise ValueError("Invalid prompt style")

    return d[prompt_name]


PROMPTS = [
    Prompt(
        "summarize",
        PromptNames.SUMMARIZE_FUNC,
        desc="Summarize the function",
        response_key="summary",
        gui_result_callback=Prompt.comment_function
    ),
    Prompt(
        "identify_source",
        PromptNames.ID_SRC,
        desc="Identify the source of the function",
        response_key="link",
        gui_result_callback=Prompt.comment_function
    ),
    Prompt(
        "rename_variables",
        PromptNames.RENAME_VARS,
        desc="Suggest variable names",
        gui_result_callback=Prompt.rename_variables
    ),
    Prompt(
        "rename_function",
        PromptNames.RENAME_FUNC,
        desc="Suggest a function name",
        gui_result_callback=Prompt.rename_function
    ),
]
